MSGLINK_VERSION = 2

AUTH_REQUEST = 0x00
AUTH_FAIL = 0x01
AUTH_OK = 0x02
PEER_ONLINE = 0x03
PEER_OFFLINE = 0x04
WATCH_ROLE = 0x05
MESSAGE_FULL = 0x06
MESSAGE_PART = 0x07
MESSAGE_END = 0x08
# PUBLISH_FULL = 0x09
# PUBLISH_PART = 0x0A
# PUBLISH_END = 0x0B
SUBSCRIBE = 0x0C
POST_FULL = 0x0D
POST_PART = 0x0E
POST_END = 0x0F
WINDOW_UPDATE = 0x10
HEARTBEAT = 0x11
DROP_BUFFERED = 0x12


MAX_MESSAGE_SIZE = 2**16 - 1
MESSAGE_FULL_MAX_PAYLOAD = MAX_MESSAGE_SIZE - 4 - 2
MESSAGE_PART_MAX_PAYLOAD = MAX_MESSAGE_SIZE - 4 - 2
POST_FULL_MAX_PAYLOAD = MAX_MESSAGE_SIZE - 4 - 32
POST_PART_MAX_PAYLOAD = MAX_MESSAGE_SIZE - 4 - 32


def _enc_str(s: str):
    if len(s) < 128:
        return len(s).to_bytes(1, 'big') + s.encode('utf-8')
    else:
        value = len(s) | (1 << 15)
        return value.to_bytes(2, 'big') + s.encode('utf-8')


def _dec_str(data: bytes, pos: int):
    # if first bit is set, length has 16 bit, else 8 bit
    if data[pos] & (1 << 7):
        str_length = int.from_bytes(data[pos:pos+2], 'big')
        str_length &= ~(1 << 15)  # unset leftmost bit
        pos += 2
    else:
        str_length = int.from_bytes(data[pos:pos+1], 'big')
        pos += 1

    return data[pos:pos+str_length].decode('utf-8'), pos + str_length


def _enc_uint8(i: int):
    return i.to_bytes(1, 'big')


def _dec_uint8(data: bytes, pos: int):
    return int.from_bytes(data[pos:pos+1], 'big'), pos + 1


def _enc_uint16(i: int):
    return i.to_bytes(2, 'big')


def _dec_uint16(data: bytes, pos: int):
    return int.from_bytes(data[pos:pos+2], 'big'), pos + 2


def _enc_uint32(i: int):
    return i.to_bytes(4, 'big')


def _dec_uint32(data, pos):
    return int.from_bytes(data[pos:pos+4], 'big'), pos + 4


def _make_packet(type: int, data: bytes):
    return _enc_uint8(type) + _enc_uint16(len(data)) + data


def encode_auth_request(role: str, password: str, peer_info: str):
    data = _enc_str(role) + _enc_str(password) + _enc_uint16(MSGLINK_VERSION)
    if peer_info is not None and len(peer_info) > 0:
        data += _enc_str(peer_info)
    return _make_packet(AUTH_REQUEST, data)


def encode_watch_role(role: str):
    return _make_packet(WATCH_ROLE, _enc_str(role))


def encode_message_full(receiver: int, channel: int, data: bytes):
    return _make_packet(MESSAGE_FULL, _enc_uint32(receiver) + _enc_uint16(channel) + data)


def encode_message_part(receiver: int, channel: int, data: bytes):
    return _make_packet(MESSAGE_PART, _enc_uint32(receiver) + _enc_uint16(channel) + data)


def encode_message_end(receiver: int, channel: int):
    return _make_packet(MESSAGE_END, _enc_uint32(receiver) + _enc_uint16(channel))


def encode_post_full(topic: bytes, data: bytes):
    return _make_packet(POST_FULL, _enc_uint32(0) + topic + data)


def encode_post_part(topic: bytes, data: bytes):
    return _make_packet(POST_PART, _enc_uint32(0) + topic + data)


def encode_post_end(topic: bytes):
    return _make_packet(POST_END, _enc_uint32(0) + topic)


def encode_subscribe(topic: bytes):
    return _make_packet(SUBSCRIBE, topic)


def encode_window_update(window_update_size: int):
    return _make_packet(WINDOW_UPDATE, _enc_uint32(window_update_size))


def encode_heartbeat():
    return _make_packet(HEARTBEAT, bytes())


def encode_drop_buffered(receiver: int, channel: int):
    return _make_packet(DROP_BUFFERED, _enc_uint32(receiver) + _enc_uint16(channel))


def decode_peer_online(data: bytes):
    peer, pos = _dec_uint32(data, 0)
    role, pos = _dec_str(data, pos)
    info = None
    if pos < len(data):
        info, pos = _dec_str(data, pos)
    return peer, role, info


def decode_peer_offline(data: bytes):
    peer, _ = _dec_uint32(data, 0)
    return peer


def decode_message_full(data: bytes):
    sender, pos = _dec_uint32(data, 0)
    channel, pos = _dec_uint16(data, pos)
    return sender, channel, data[pos:]


def decode_message_part(data: bytes):
    sender, pos = _dec_uint32(data, 0)
    channel, pos = _dec_uint16(data, pos)
    return sender, channel, data[pos:]


def decode_message_end(data: bytes):
    sender, pos = _dec_uint32(data, 0)
    channel, pos = _dec_uint16(data, pos)
    return sender, channel


def decode_post_full(data: bytes):
    sender, pos = _dec_uint32(data, 0)
    topic_id = data[pos:pos+32]
    return sender, bytes(topic_id), data[pos+32:]


def decode_post_part(data: bytes):
    sender, pos = _dec_uint32(data, 0)
    topic_id = data[pos:pos+32]
    return sender, bytes(topic_id), data[pos+32:]


def decode_post_end(data: bytes):
    sender, pos = _dec_uint32(data, 0)
    topic_id = data[pos:pos+32]
    return sender, bytes(topic_id)


def decode_window_update(data: bytes):
    window_update, _ = _dec_uint32(data, 0)
    return window_update


def decode_drop_buffered(data: bytes):
    peer, pos = _dec_uint32(data, 0)
    channel, pos = _dec_uint16(data, pos)
    return peer, channel



import msglink
from typing import Dict, Set
from dataclasses import dataclass
import asyncio
import ssl
import logging
import socket
from contextlib import suppress

log = logging.getLogger()


class Router(msglink.Callbacks):
    def __init__(self, server, port, role, password, info=""):
        self._protocol = msglink.Protocol(self)
        self._connected_peers: Dict[int, Router.Peer] = dict()
        self._server = server
        self._port = port
        self._role = role
        self._password = password
        self._info = info
        self._reconnect_interval = 10
        self._running = False
        self._peer_discoveries: Set[Router.PeerDiscovery] = set()
        self._disconnect_waits: Set[Router.DisconnectWait] = set()
        self._message_readers: Set[Router.MessgeReader] = set()
        self._client_readers: Set[Router.ClientReader] = set()
        self._channel_readers: Set[Router.ChannelReader] = set()
        self._topic_readers: Set[Router.TopicReader] = set()
        self._established = False
        self._reconnect_task: asyncio.Task | None = None
        self._auth_future: asyncio.Future | None = None

    def start(self):
        if self._running:
            return

        self._running = True
        self._reconnect_task = asyncio.create_task(self._reconnect_loop())

    async def stop(self):
        self._running = False
        self._reconnect_task.cancel()
        with suppress(asyncio.CancelledError):
            await self._reconnect_task
        await self._protocol.async_close()

    def set_reconnect_interval(self, reconnect_interval: int):
        self._reconnect_interval = reconnect_interval

    async def discover_one(self, role: str):
        for pid, peer in self._connected_peers.items():
            if peer.role == role:
                return pid

        disc = Router.PeerDiscovery(role=role)
        self._peer_discoveries.add(disc)
        if self._established:
            self._protocol.send_watch_role(role)
        pid = await disc.queue.get()
        disc.queue.task_done()
        self._peer_discoveries.remove(disc)
        return pid

    async def discover_all(self, role: str):
        disc = Router.PeerDiscovery(role=role)
        self._peer_discoveries.add(disc)

        if self._established:
            self._protocol.send_watch_role(role)

        # store known and matching in list before yielding to avoid race conditions
        # otherwise _known_peers could change mid-iteration during the yield
        matching_known = list()
        for pid, peer in self._connected_peers.items():
            if peer.role == role:
                matching_known.append(pid)

        try:
            for pid in matching_known:
                yield pid

            while self._running:
                yield await disc.queue.get()
                disc.queue.task_done()

        finally:  # handles GeneratorExit
            self._peer_discoveries.remove(disc)

    async def wait_peer_disconnected(self, peer: int):
        if peer not in self._connected_peers:
            return

        wait = Router.DisconnectWait(peer)
        self._disconnect_waits.add(wait)
        await wait.future
        self._disconnect_waits.remove(wait)

    async def read_messages(self):
        mr = Router.MessgeReader()
        self._message_readers.add(mr)

        try:
            while self._running:
                msg = await mr.queue.get()
                if msg is not None:
                    yield msg
                else:
                    return
        finally:
            self._message_readers.remove(mr)

    async def read_peer_messages(self, peer: int):
        if peer not in self._connected_peers:
            return

        cr = Router.ClientReader(peer)
        self._client_readers.add(cr)

        try:
            while self._running:
                msg = await cr.queue.get()
                if msg is not None:
                    yield msg
                else:
                    return
        finally:
            self._client_readers.remove(cr)

    async def read_peer_channel(self, peer: int, channel: int):
        if peer not in self._connected_peers:
            return

        cr = Router.ChannelReader(peer, channel)
        self._channel_readers.add(cr)

        try:
            while self._running:
                msg = await cr.queue.get()
                if msg is not None:
                    yield msg
                else:
                    return
        finally:
            self._channel_readers.remove(cr)

    def get_peer_role(self, peer: int) -> str | None:
        if peer not in self._connected_peers:
            return None
        else:
            return self._connected_peers[peer].role

    def get_peer_info(self, peer: int) -> str | None:
        if peer not in self._connected_peers:
            return None
        else:
            return self._connected_peers[peer].info

    async def subscribe_topic(self, topic: str):
        sub = Router.TopicReader(topic=topic)
        self._topic_readers.add(sub)
        self._protocol.send_subscribe(topic)

        try:
            while self._running:
                msg = await sub.queue.get()
                if msg is not None:
                    yield msg
                else:
                    return
        finally:
            self._topic_readers.remove(sub)

    async def send_message(self, receiver: int, channel: int, msg: bytes):
        await self._protocol.send_message(receiver, channel, msg)

    async def make_post(self, topic: str, msg: bytes):
        await self._protocol.send_post(topic, msg)

    # ==================================================================================================================
    # end of public interface
    # ==================================================================================================================

    @dataclass
    class Peer:
        role: str
        info: str

    class PeerDiscovery:
        def __init__(self, role):
            self.role = role
            self.queue = asyncio.Queue()

    class DisconnectWait:
        def __init__(self, peer):
            self.pid = peer
            self.future = asyncio.Future()

    class MessgeReader:
        def __init__(self):
            self.queue = asyncio.Queue()

    class ClientReader:
        def __init__(self, peer: int):
            self.pid = peer
            self.queue = asyncio.Queue()

    class ChannelReader:
        def __init__(self, peer: int, channel: int):
            self.pid = peer
            self.channel = channel
            self.queue = asyncio.Queue()

    class TopicReader:
        def __init__(self, topic: str):
            self.topic = topic
            self.queue = asyncio.Queue()

    async def _try_connect_msglink(self):
        if self._protocol.is_connected():
            # tcp connection already established
            return

        log.info(f"[msglink] connecting to {self._server}:{self._port}")
        loop = asyncio.get_running_loop()
        await loop.create_connection(lambda: self._protocol, self._server, self._port, ssl=ssl.SSLContext())
        log.info(f"[msglink] connected to {self._server}:{self._port}")

    async def _try_authenticate(self):
        if self._auth_future is not None and self._auth_future.result():
            # authentication already established
            return

        # need to make new authentication attempt
        # (_auth_result either None or failed)
        log.info(f"[msglink] authenticating to {self._server}")
        self._auth_future = asyncio.Future()
        self._protocol.send_auth_request(self._role, self._password, self._info)
        await self._auth_future

    async def _reconnect_loop(self):
        while self._running:
            try:
                await self._try_connect_msglink()
                await self._try_authenticate()

            except socket.error as e:
                log.warning(f"[msglink] error connecting to {self._server}: {e}")
            await asyncio.sleep(self._reconnect_interval)

    def _all_disconnected(self):
        for d in self._disconnect_waits:
            d.future.set_result(None)

        for cr in self._channel_readers:
            cr.queue.put_nowait(None)

    # msglink callbacks =================================================================
    def on_connect_close(self):
        log.info(f"[msglink] connection to {self._server}:{self._port} was closed")
        self._established = False
        self._connected_peers.clear()
        self._all_disconnected()
        self._auth_future = None

    def on_auth_fail(self):
        log.warning(f"[msglink] authentication failed for role '{self._role}' at server {self._server}")
        self._auth_future.set_result(False)

    def on_auth_success(self):
        log.info(f"[msglink] authentication successful")
        self._established = True
        self._auth_future.set_result(True)
        for pd in self._peer_discoveries:
            self._protocol.send_watch_role(pd.role)

    def on_peer_online(self, peer, role, info):
        self._connected_peers[peer] = Router.Peer(role=role, info=info)
        for pd in self._peer_discoveries:
            if pd.role == role:
                pd.queue.put_nowait(peer)

    def on_peer_offline(self, peer, role):
        del self._connected_peers[peer]
        for d in self._disconnect_waits:
            if d.pid == peer:
                d.future.set_result(None)

        for cr in self._channel_readers:
            if cr.pid == peer:
                cr.queue.put_nowait(None)

    def on_message(self, sender, role, channel, payload):
        for mr in self._message_readers:
            mr.queue.put_nowait((sender, channel, payload))

        for cr in self._client_readers:
            if cr.pid == sender:
                cr.queue.put_nowait((channel, payload))

        for cr in self._channel_readers:
            if cr.pid == sender and cr.channel == channel:
                cr.queue.put_nowait(payload)

    def on_post(self, sender, role, topic, payload):
        for sub in self._topic_readers:
            if sub.topic == topic:
                sub.queue.put_nowait(payload)



from setuptools import setup

setup(
    name='msglink',
    version='1.5.6',
    packages=['msglink'],
    url='',
    license='LGPL v3',
    author='Robert Brünig',
    author_email='',
    description='msglink client library',
    package_data={"msglink": ["py.typed"]},
)



import asyncio
from typing import List, Dict, Optional
from collections import deque, defaultdict
from msglink import packets
import hashlib
import time
from contextlib import suppress



BUFFER_TIME_MS = 500
MIN_WINDOW = 3000
UINT32_MAX = 2**32 - 1


class Callbacks:
    def on_connect_close(self):
        pass

    def on_auth_fail(self):
        pass

    def on_auth_success(self):
        pass

    def on_peer_online(self, peer, role, info):
        pass

    def on_peer_offline(self, peer, role):
        pass

    def on_message(self, sender, role, channel, payload):
        pass

    def on_post(self, sender, role, topic, payload):
        pass


class Stream:
    def __init__(self, receiver, channel, topic):
        self.queue = deque()
        self.bytes_sent = 0
        self.receiver = receiver
        self.channel = channel
        self.topic = topic


class Protocol(asyncio.Protocol):
    def __init__(self, callbacks: Callbacks):
        self._callbacks = callbacks
        self._send_window = MIN_WINDOW
        self._rec_window = MIN_WINDOW
        self._bytes_since_update = 0
        self._last_update_time = time.perf_counter()
        self._streams: List[Stream] = list()
        self._transport: Optional[asyncio.WriteTransport] = None
        self._input_buffer = bytearray()
        self._is_connected = False
        self._peer_roles = dict()
        self._heartbeat_task: Optional[asyncio.Task] = None
        self._message_buffers: Dict[tuple, bytearray] = defaultdict(bytearray)
        self._post_buffers: Dict[tuple, bytearray] = defaultdict(bytearray)
        self._topics: Dict[bytes, str] = dict()
        self._send_paused = False
        self._debug_flow = False

    def pause_writing(self) -> None:
        self._send_paused = True

    def resume_writing(self) -> None:
        self._send_paused = False
        self._send_queued()

    def _clear(self):
        self._send_window = MIN_WINDOW
        self._rec_window = MIN_WINDOW
        self._bytes_since_update = 0
        self._last_update_time = time.perf_counter()
        self._streams.clear()
        self._input_buffer = bytearray()
        self._is_connected = False
        self._peer_roles.clear()
        if self._heartbeat_task:
            self._heartbeat_task.cancel()
            self._heartbeat_task = None

    def connection_made(self, transport: asyncio.WriteTransport):
        self._clear()
        self._transport = transport
        self._is_connected = True
        self._heartbeat_task = asyncio.ensure_future(self._heartbeat_loop())

    def data_received(self, data):
        self._input_buffer.extend(data)
        while len(self._input_buffer) >= 3:
            packet_type = int.from_bytes(self._input_buffer[0:1], 'big')
            payload_size = int.from_bytes(self._input_buffer[1:3], 'big')
            if len(self._input_buffer) < 3 + payload_size:
                return
            payload = self._input_buffer[3:3+payload_size]
            self._input_buffer = self._input_buffer[3+payload_size:]
            self._handle_packet(packet_type, payload)

    def connection_lost(self, exc):
        self._is_connected = False
        self._clear()
        self._callbacks.on_connect_close()

    def close(self):
        self._transport.abort()

    async def async_close(self):
        unfinished = self._heartbeat_task
        self.close()
        if unfinished:
            with suppress(asyncio.CancelledError):
                await unfinished

    def is_connected(self):
        return self._is_connected

    def send_auth_request(self, role, password, info):
        self._send_control_packet(packets.encode_auth_request(role, password, info))

    def send_watch_role(self, role):
        self._send_control_packet(packets.encode_watch_role(role))

    async def send_message(self, receiver: int, channel: int, payload: bytes):
        if len(payload) <= packets.MESSAGE_FULL_MAX_PAYLOAD:
            packet = packets.encode_message_full(receiver, channel, payload)
            await self._send_message_packet(receiver, channel, None, packet)
        else:
            pos = 0
            while pos < len(payload) - 1:
                size = min(len(payload) - pos, packets.MESSAGE_PART_MAX_PAYLOAD)
                packet = packets.encode_message_part(receiver, channel, payload[pos:pos+size])
                await self._send_message_packet(receiver, channel, None, packet)
                pos += size
            packet = packets.encode_message_end(receiver, channel)
            await self._send_message_packet(receiver, channel, None, packet)

    async def send_post(self, topic: str, payload: bytes):
        topic_id = hashlib.sha256(topic.encode("utf-8")).digest()
        if len(payload) <= packets.POST_FULL_MAX_PAYLOAD:
            packet = packets.encode_post_full(topic_id, payload)
            await self._send_message_packet(None, None, topic, packet)
        else:
            pos = 0
            while pos < len(payload) - 1:
                size = min(len(payload) - pos, packets.POST_PART_MAX_PAYLOAD)
                packet = packets.encode_post_part(topic_id, payload[pos:pos + size])
                await self._send_message_packet(None, None, topic, packet)
                pos += size
            packet = packets.encode_post_end(topic_id)
            await self._send_message_packet(None, None, topic, packet)

    def send_subscribe(self, topic: str):
        topic_id = hashlib.sha256(topic.encode('utf-8')).digest()
        self._topics[topic_id] = topic
        self._send_control_packet(packets.encode_subscribe(topic_id))

    def send_drop_buffered(self, receiver, channel):
        self._send_control_packet(packets.encode_drop_buffered(receiver, channel))

    def set_debug_flow(self, debug):
        self._debug_flow = debug

    def _send_control_packet(self, packet):
        self._transport.write(packet)

    async def _send_message_packet(self, receiver, channel, topic, packet):
        if self._send_window <= 0 or self._send_paused:
            stream = self._get_stream(receiver, channel, topic)
            future = asyncio.Future()
            stream.queue.append((future, packet))
            await future
        else:
            self._transport.write(packet)
            self._send_window -= len(packet) - 3

    def _get_stream(self, receiver, channel, topic):
        for stream in self._streams:
            if stream.receiver == receiver and stream.channel == channel and stream.topic == topic:
                return stream
        stream = Stream(receiver, channel, topic)
        self._streams.append(stream)
        return stream

    def _handle_packet(self, packet_type, packet_data):
        self._extend_deadline = True
        if packet_type == packets.AUTH_FAIL:
            self._callbacks.on_auth_fail()

        elif packet_type == packets.AUTH_OK:
            self._callbacks.on_auth_success()

        elif packet_type == packets.PEER_ONLINE:
            peer, role, info = packets.decode_peer_online(packet_data)
            self._peer_roles[peer] = role
            self._callbacks.on_peer_online(peer, role, info)

        elif packet_type == packets.PEER_OFFLINE:
            peer = packets.decode_peer_offline(packet_data)
            role = self._peer_roles.pop(peer)
            self._callbacks.on_peer_offline(peer, role)

        elif packet_type == packets.MESSAGE_FULL:
            sender, channel, payload = packets.decode_message_full(packet_data)
            self._update_receive_flow(len(packet_data))
            role = self._peer_roles[sender]
            self._callbacks.on_message(sender, role, channel, payload)

        elif packet_type == packets.MESSAGE_PART:
            sender, channel, payload = packets.decode_message_part(packet_data)
            self._update_receive_flow(len(packet_data))
            self._message_buffers[(sender, channel)].extend(payload)

        elif packet_type == packets.MESSAGE_END:
            sender, channel = packets.decode_message_end(packet_data)
            self._update_receive_flow(len(packet_data))
            payload = self._message_buffers.pop((sender, channel))
            role = self._peer_roles[sender]
            self._callbacks.on_message(sender, role, channel, payload)

        elif packet_type == packets.POST_FULL:
            sender, topic_id, payload = packets.decode_post_full(packet_data)
            self._update_receive_flow(len(packet_data))
            role = self._peer_roles[sender]
            topic = self._topics.get(topic_id, None)
            self._callbacks.on_post(sender, role, topic, payload)

        elif packet_type == packets.POST_PART:
            sender, topic_id, payload = packets.decode_post_part(packet_data)
            self._update_receive_flow(len(packet_data))
            self._post_buffers[(sender, topic_id)].extend(payload)

        elif packet_type == packets.POST_END:
            sender, topic_id = packets.decode_post_end(packet_data)
            self._update_receive_flow(len(packet_data))
            payload = self._post_buffers.pop((sender, topic_id))
            role = self._peer_roles[sender]
            topic = self._topics.get(topic_id, None)
            self._callbacks.on_post(sender, role, topic, payload)

        elif packet_type == packets.WINDOW_UPDATE:
            self._send_window += packets.decode_window_update(packet_data)
            self._send_queued()

        elif packet_type == packets.HEARTBEAT:
            pass

        elif packet_type == packets.DROP_BUFFERED:
            peer, channel = packets.decode_drop_buffered(packet_data)
            self._message_buffers.pop((peer, channel), None)

        else:
            print(f"got packet with invalid type {packet_type}")

    def _send_queued(self):
        if not self._streams or self._send_window <= 0 or self._send_paused:
            return

        # tare bytes_sent
        min_stream = min(self._streams, key=lambda x: x.bytes_sent)
        min_sent = min_stream.bytes_sent
        for stream in self._streams:
            stream.bytes_sent -= min_sent

        # send queued
        while self._send_window > 0 and self._streams and not self._send_paused:
            index, stream = min(enumerate(self._streams), key=lambda x: x[1].bytes_sent)
            future, packet = stream.queue.popleft()
            stream.bytes_sent += len(packet) - 3
            self._transport.write(packet)
            self._send_window -= len(packet) - 3
            future.set_result(None)
            if not stream.queue:
                del self._streams[index]

    def _update_receive_flow(self, received_bytes):
        # (1) update count of received bytes
        self._bytes_since_update += received_bytes

        # (2) Is the number of bytes received smaller than a quarter of the current window?
        if self._bytes_since_update < self._rec_window / 4:
            return

        # (3) Calculate new window size. new_window = buffer_time / measured_time * received_bytes (+ apply bounds)
        now = time.perf_counter()
        ms_since_last_update = max(round((now - self._last_update_time) * 1000), 1)
        new_window = round(BUFFER_TIME_MS / ms_since_last_update * self._bytes_since_update)
        min_new_window = max(MIN_WINDOW, round(self._rec_window / 2))
        max_new_window = min(UINT32_MAX, self._rec_window * 2)
        new_window = max(min_new_window, min(new_window, max_new_window))

        if self._debug_flow:
            old_speed = _fmt_size(self._rec_window * 1000 / BUFFER_TIME_MS)
            new_speed = _fmt_size(new_window * 1000 / BUFFER_TIME_MS)
            print(f"speed {old_speed}/s -> {new_speed}/s  "
                  f"window {_fmt_size(self._rec_window)} -> {_fmt_size(new_window)}  "
                  f"received {_fmt_size(self._bytes_since_update)} in {ms_since_last_update}ms")

        # (4) Calculate window update size
        window_update = new_window + self._bytes_since_update - self._rec_window

        # (5) Is the window update size positive (could be negative when window is shrinking
        if window_update <= 0:
            return

        # (6) Save state and send window update
        self._last_update_time = now
        self._rec_window = new_window
        self._bytes_since_update = 0
        self._send_control_packet(packets.encode_window_update(window_update))

    async def _heartbeat_loop(self):
        try:
            while True:
                # send heartbeat every 30s, check connection every 60s
                await asyncio.sleep(20)
                self._send_control_packet(packets.encode_heartbeat())
                await asyncio.sleep(20)
                self._send_control_packet(packets.encode_heartbeat())
                await asyncio.sleep(20)
                self._send_control_packet(packets.encode_heartbeat())
                if self._extend_deadline:
                    self._extend_deadline = False
                else:
                    self._transport.abort()
                    return
        except asyncio.CancelledError:
            pass


def _fmt_size(size):
    for suffix in ["Byte", "KB", "MB", "GB"]:
        if size < 1024:
            break
        size /= 1024
    return f"{round(size)} {suffix}"
